@@ -39,14 +39,7 @@ use PublicInbox::DS qw(now);
 use PublicInbox::GitAsyncCat;
 use Text::ParseWords qw(parse_line);
 use Errno qw(EAGAIN);
-use PublicInbox::IMAPsearchqp;
-
-my $Address;
-for my $mod (qw(Email::Address::XS Mail::Address)) {
-	eval "require $mod" or next;
-	$Address = $mod and last;
-}
-die "neither Email::Address::XS nor Mail::Address loaded: $@" if !$Address;
+use PublicInbox::Address;
 
 sub LINE_MAX () { 8000 } # RFC 2683 3.2.1.5
 
@@ -95,7 +88,7 @@ for my $att (keys %FETCH_ATT) {
 }
 undef %FETCH_NEED;
 
-my $valid_range = '[0-9]+|[0-9]+:[0-9]+|[0-9]+:\*';
+my $valid_range = '[0-9]+|[0-9]+:[0-9]+|[0-9]+:\*|\*';
 $valid_range = qr/\A(?:$valid_range)(?:,(?:$valid_range))*\z/;
 
 sub do_greet {
@@ -105,8 +98,9 @@ sub do_greet {
 }
 
 sub new {
-	my (undef, $sock, $imapd) = @_;
-	(bless { imapd => $imapd }, 'PublicInbox::IMAP_preauth')->greet($sock)
+	my (undef, $sock, $addr, $imapd) = @_;
+	(bless { imapd => $imapd }, 'PublicInbox::IMAP_preauth')
+		->greet($sock, $addr)
 }
 
 sub logged_in { 1 }
@@ -158,14 +152,14 @@ sub cmd_logout ($$) {
 	my ($self, $tag) = @_;
 	delete $self->{-idle_tag};
 	$self->write(\"* BYE logging out\r\n$tag OK Logout done\r\n");
-	$self->shutdn; # PublicInbox::DS::shutdn
+	$self->close;
 	undef;
 }
 
 sub cmd_authenticate ($$$) {
 	my ($self, $tag) = @_; # $method = $_[2], should be "ANONYMOUS"
 	$self->{-login_tag} = $tag;
-	"+\r\n"; # challenge
+	"+ \r\n"; # challenge
 }
 
 sub cmd_capability ($$) {
@@ -426,8 +420,10 @@ sub _esc ($) {
 	if (!defined($v)) {
 		'NIL';
 	} elsif ($v =~ /[{"\r\n%*\\\[]/) { # literal string
+		utf8::encode($v);
 		'{' . length($v) . "}\r\n" . $v;
 	} else { # quoted string
+		utf8::encode($v);
 		qq{"$v"}
 	}
 }
@@ -437,7 +433,7 @@ sub addr_envelope ($$;$) {
 	my $v = $eml->header_raw($x) //
 		($y ? $eml->header_raw($y) : undef) // return 'NIL';
 
-	my @x = $Address->parse($v) or return 'NIL';
+	my @x = PublicInbox::Address::objects($v) or return 'NIL';
 	'(' . join('',
 		map { '(' . join(' ',
 				_esc($_->name), 'NIL',
@@ -590,14 +586,16 @@ sub fetch_blob_cb { # called by git->cat_async via ibx_async_cat
 	my ($self, undef, $msgs, $range_info, $ops, $partial) = @$fetch_arg;
 	my $ibx = $self->{ibx} or return $self->close; # client disconnected
 	my $smsg = shift @$msgs or die 'BUG: no smsg';
-	if (!defined($oid)) {
+	if (!defined($type)) {
+		warn "E: git aborted on $oid / $smsg->{blob} $ibx->{inboxdir}";
+		return $self->close;
+	} elsif ($type ne 'blob') {
 		# it's possible to have TOCTOU if an admin runs
 		# public-inbox-(edit|purge), just move onto the next message
-		warn "E: $smsg->{blob} missing in $ibx->{inboxdir}\n";
+		warn "E: $smsg->{blob} $type in $ibx->{inboxdir}\n";
 		return $self->requeue_once;
-	} else {
-		$smsg->{blob} eq $oid or die "BUG: $smsg->{blob} != $oid";
 	}
+	$smsg->{blob} eq $oid or die "BUG: $smsg->{blob} != $oid";
 	my $pre;
 	($self->{anon} && !$self->{wbuf} && $msgs->[0]) and
 		$pre = ibx_async_prefetch($ibx, $msgs->[0]->{blob},
@@ -608,8 +606,7 @@ sub fetch_blob_cb { # called by git->cat_async via ibx_async_cat
 
 sub emit_rfc822 {
 	my ($self, $k, undef, $bref) = @_;
-	$self->msg_more(" $k {" . length($$bref)."}\r\n");
-	$self->msg_more($$bref);
+	$self->msg_more(" $k {" . length($$bref)."}\r\n", $$bref);
 }
 
 # Mail::IMAPClient::message_string cares about this by default,
@@ -629,20 +626,18 @@ sub emit_flags { $_[0]->msg_more(' FLAGS ()') }
 
 sub emit_envelope {
 	my ($self, undef, undef, undef, $eml) = @_;
-	$self->msg_more(' ENVELOPE '.eml_envelope($eml));
+	$self->msg_more(' ENVELOPE ', eml_envelope($eml));
 }
 
 sub emit_rfc822_header {
 	my ($self, $k, undef, undef, $eml) = @_;
-	$self->msg_more(" $k {".length(${$eml->{hdr}})."}\r\n");
-	$self->msg_more(${$eml->{hdr}});
+	$self->msg_more(" $k {".length(${$eml->{hdr}})."}\r\n", ${$eml->{hdr}});
 }
 
 # n.b. this is sorted to be after any emit_eml_new ops
 sub emit_rfc822_text {
 	my ($self, $k, undef, $bref) = @_;
-	$self->msg_more(" $k {".length($$bref)."}\r\n");
-	$self->msg_more($$bref);
+	$self->msg_more(" $k {".length($$bref)."}\r\n", $$bref);
 }
 
 sub emit_bodystructure {
@@ -661,7 +656,7 @@ sub op_eml_new { $_[4] = PublicInbox::Eml->new($_[3]) }
 # s/From / fixes old bug from import (pre-a0c07cba0e5d8b6a)
 sub to_crlf_full {
 	${$_[0]} =~ s/(?<!\r)\n/\r\n/sg;
-	${$_[0]} =~ s/\A[\r\n]*From [^\r\n]*\r\n//s;
+	PublicInbox::Eml::strip_from(${$_[0]});
 }
 
 sub op_crlf_bref { to_crlf_full($_[3]) }
@@ -703,6 +698,9 @@ sub range_step ($$) {
 		# just let the caller do an out-of-range query if a single
 		# UID is out-of-range
 		++$beg if ($beg <= $uid_base || $end > $uid_end);
+	} elsif ($range eq '*') {
+		$beg = $end = $self->{ibx}->over(1)->max;
+		uid_clamp($self, \$beg, \$end);
 	} else {
 		return 'BAD fetch range';
 	}
@@ -973,8 +971,7 @@ sub partial_emit ($$$) {
 		} else {
 			$len = length($str);
 		}
-		$self->msg_more(" $k {$len}\r\n");
-		$self->msg_more($str);
+		$self->msg_more(" $k {$len}\r\n", $str);
 	}
 }
 
@@ -1000,12 +997,12 @@ sub fetch_compile ($) {
 			return "BAD param: $att";
 		}
 	}
-	my @r;
+	my ($cb, $ops, $partial); # return values
 
 	# stabilize partial order for consistency and ease-of-debugging:
 	if (scalar keys %partial) {
 		$need |= NEED_BLOB;
-		@{$r[2]} = map { [ $_, @{$partial{$_}} ] } sort keys %partial;
+		@$partial = map { [ $_, @{$partial{$_}} ] } sort keys %partial;
 	}
 
 	push @op, $OP_EML_NEW if ($need & (EML_HDR|EML_BDY));
@@ -1023,13 +1020,13 @@ sub fetch_compile ($) {
 		}
 	}
 
-	$r[0] = $need & NEED_BLOB ? \&fetch_blob :
+	$cb = $need & NEED_BLOB ? \&fetch_blob :
 		($need & NEED_SMSG ? \&fetch_smsg : \&fetch_uid);
 
 	# r[1] = [ $key1, $cb1, $key2, $cb2, ... ]
 	use sort 'stable'; # makes output more consistent
-	@{$r[1]} = map { ($_->[2], $_->[1]) } sort { $a->[0] <=> $b->[0] } @op;
-	@r;
+	@$ops = map { ($_->[2], $_->[1]) } sort { $a->[0] <=> $b->[0] } @op;
+	($cb, $ops, $partial);
 }
 
 sub cmd_uid_fetch ($$$$;@) {
@@ -1084,6 +1081,7 @@ sub search_uid_range { # long_response
 
 sub parse_imap_query ($$) {
 	my ($self, $query) = @_;
+	# IMAPsearchqp gets loaded in IMAPD->refresh_groups
 	my $q = PublicInbox::IMAPsearchqp::parse($self, $query);
 	if (ref($q)) {
 		my $max = $self->{ibx}->over(1)->max;
@@ -1094,6 +1092,18 @@ sub parse_imap_query ($$) {
 	$q;
 }
 
+sub mset_cb { # async_mset cb
+	my ($self, $tag, $want_msn, $srch, $opt, $mset, $err) = @_;
+	$self->write($err ? do {
+		warn "IMAP search error: $err\n";
+		\"$tag BAD program fault - command not performed\r\n"
+	} : do {
+		my $uids = $srch->mset_to_artnums($mset, $opt);
+		msn_convert($self, $uids) if scalar(@$uids) && $want_msn;
+		\"* SEARCH @$uids\r\n$tag OK Search done\r\n"
+	});
+}
+
 sub search_common {
 	my ($self, $tag, $query, $want_msn) = @_;
 	my $ibx = $self->{ibx} or return "$tag BAD No mailbox selected\r\n";
@@ -1108,14 +1118,14 @@ sub search_common {
 		my $srch = $self->{ibx}->isrch or
 			return "$tag BAD search not available for mailbox\r\n";
 		my $opt = {
-			relevance => -1,
+			sort_col => -1,
+			asc => 1,
 			limit => UID_SLICE,
 			uid_range => $range_info
 		};
-		my $mset = $srch->mset($q, $opt);
-		my $uids = $srch->mset_to_artnums($mset, $opt);
-		msn_convert($self, $uids) if scalar(@$uids) && $want_msn;
-		"* SEARCH @$uids\r\n$tag OK Search done\r\n";
+		$srch->async_mset($q, $opt, \&mset_cb,
+					$self, $tag, $want_msn, $srch, $opt) ?
+				undef : '';
 	} else {
 		"$tag BAD Error\r\n";
 	}
