@@ -15,10 +15,11 @@ use PublicInbox::MID qw(mid_escape $MID_EXTRACT);
 use PublicInbox::Eml;
 use POSIX qw(strftime);
 use PublicInbox::DS qw(now);
-use Digest::SHA qw(sha1_hex);
+use PublicInbox::SHA qw(sha1_hex);
 use Time::Local qw(timegm timelocal);
 use PublicInbox::GitAsyncCat;
 use PublicInbox::Address;
+my $SALT = rand . ''; # stringify early for CoW sharing
 
 use constant {
 	LINE_MAX => 512, # RFC 977 section 2.3
@@ -26,8 +27,8 @@ use constant {
 	r502 => "502 Command unavailable\r\n",
 	r221 => "221 Header follows\r\n",
 	r225 =>	"225 Headers follow (multi-line)\r\n",
-	r430 => "430 No article with that message-id\r\n",
 };
+sub r430 ($) { "430 No article with that message-id <$_[0]>\r\n" }
 use Errno qw(EAGAIN);
 my $ONE_MSGID = qr/\A$MID_EXTRACT\z/;
 my @OVERVIEW = qw(Subject From Date Message-ID References);
@@ -48,8 +49,8 @@ COMPRESS DEFLATE\r
 sub do_greet ($) { $_[0]->write($_[0]->{nntpd}->{greet}) };
 
 sub new {
-	my ($cls, $sock, $nntpd) = @_;
-	(bless { nntpd => $nntpd }, $cls)->greet($sock)
+	my ($cls, $sock, $addr, $nntpd) = @_;
+	(bless { nntpd => $nntpd }, $cls)->greet($sock, $addr)
 }
 
 sub args_ok ($$) {
@@ -119,7 +120,14 @@ sub names2ibx ($;$) {
 	}
 }
 
-sub list_active_i { # "LIST ACTIVE" and also just "LIST" (no args)
+sub _list_groups (@) {
+	my ($cb, $self, $wildmat) = @_;
+	my $re = wildmat2re($wildmat);
+	my @names = grep /$re/, @{$self->{nntpd}->{groupnames}};
+	$self->long_response($cb, names2ibx($self, \@names));
+}
+
+sub _list_active_i { # "LIST ACTIVE" and also just "LIST" (no args)
 	my ($self, $ibxs) = @_;
 	my @window = splice(@$ibxs, 0, 1000);
 	emit_group_lines($self, \@window);
@@ -127,13 +135,10 @@ sub list_active_i { # "LIST ACTIVE" and also just "LIST" (no args)
 }
 
 sub list_active ($;$) { # called by cmd_list
-	my ($self, $wildmat) = @_;
-	wildmat2re($wildmat);
-	my @names = grep(/$wildmat/, @{$self->{nntpd}->{groupnames}});
-	$self->long_response(\&list_active_i, names2ibx($self, \@names));
+	_list_groups \&_list_active_i, @_;
 }
 
-sub list_active_times_i {
+sub _list_active_times_i {
 	my ($self, $ibxs) = @_;
 	my @window = splice(@$ibxs, 0, 1000);
 	$self->msg_more(join('', map {
@@ -144,13 +149,10 @@ sub list_active_times_i {
 }
 
 sub list_active_times ($;$) { # called by cmd_list
-	my ($self, $wildmat) = @_;
-	wildmat2re($wildmat);
-	my @names = grep(/$wildmat/, @{$self->{nntpd}->{groupnames}});
-	$self->long_response(\&list_active_times_i, names2ibx($self, \@names));
+	_list_groups \&_list_active_times_i, @_;
 }
 
-sub list_newsgroups_i {
+sub _list_newsgroups_i {
 	my ($self, $ibxs) = @_;
 	my @window = splice(@$ibxs, 0, 1000);
 	$self->msg_more(join('', map {
@@ -160,10 +162,7 @@ sub list_newsgroups_i {
 }
 
 sub list_newsgroups ($;$) { # called by cmd_list
-	my ($self, $wildmat) = @_;
-	wildmat2re($wildmat);
-	my @names = grep(/$wildmat/, @{$self->{nntpd}->{groupnames}});
-	$self->long_response(\&list_newsgroups_i, names2ibx($self, \@names));
+	_list_groups \&_list_newsgroups_i, @_;
 }
 
 # LIST SUBSCRIPTIONS, DISTRIB.PATS are not supported
@@ -180,7 +179,7 @@ sub cmd_list ($;$$) {
 		$arg->($self, @args);
 	} else {
 		$self->msg_more("215 list of newsgroups follows\r\n");
-		$self->long_response(\&list_active_i, names2ibx($self));
+		$self->long_response(\&_list_active_i, names2ibx($self));
 	}
 }
 
@@ -188,7 +187,7 @@ sub listgroup_range_i {
 	my ($self, $beg, $end) = @_;
 	my $r = $self->{ibx}->mm(1)->msg_range($beg, $end, 'num');
 	scalar(@$r) or return;
-	$self->msg_more(join("\r\n", @$r, ''));
+	$self->msg_more(join('', map { "$_->[0]\r\n" } @$r));
 	1;
 }
 
@@ -276,15 +275,14 @@ sub cmd_newgroups ($$$;$$) {
 	$self->long_response(\&newgroups_i, $ts, names2ibx($self));
 }
 
-sub wildmat2re (;$) {
-	return $_[0] = qr/.*/ if (!defined $_[0] || $_[0] eq '*');
+sub wildmat2re ($) {
+	my $tmp = $_[0] // '*';
+	return qr/.*/ if $tmp eq '*';
 	my %keep;
-	my $salt = rand;
-	my $tmp = $_[0];
 
 	$tmp =~ s#(?<!\\)\[(.+)(?<!\\)\]#
 		my $orig = $1;
-		my $key = sha1_hex($orig . $salt);
+		my $key = sha1_hex($orig . $SALT);
 		$orig =~ s/([^\w\-])+/\Q$1/g;
 		$keep{$key} = $orig;
 		$key
@@ -297,14 +295,14 @@ sub wildmat2re (;$) {
 			defined $orig ? $orig : $1;
 			#ge;
 	}
-	$_[0] = qr/\A$tmp\z/;
+	qr/\A$tmp\z/;
 }
 
-sub ngpat2re (;$) {
-	return $_[0] = qr/\A\z/ unless defined $_[0];
+sub ngpat2re ($) {
+	my $tmp = $_[0] // return qr/\A\z/;
 	my %map = ('*' => '.*', ',' => '|');
-	$_[0] =~ s!(.)!$map{$1} || "\Q$1"!ge;
-	$_[0] = qr/\A(?:$_[0])\z/;
+	$tmp =~ s!(.)!$map{$1} || "\Q$1\E"!ge;
+	qr/\A(?:$tmp)\z/;
 }
 
 sub newnews_i {
@@ -334,8 +332,8 @@ sub cmd_newnews ($$$$;$$) {
 	return r501 if $@;
 	$self->msg_more("230 list of new articles by message-id follows\r\n");
 	my ($keep, $skip) = split(/!/, $newsgroups, 2);
-	ngpat2re($keep);
-	ngpat2re($skip);
+	$keep = ngpat2re $keep;
+	$skip = ngpat2re $skip;
 	my @names = grep(/$keep/, @{$self->{nntpd}->{groupnames}});
 	@names = grep(!/$skip/, @names);
 	return \".\r\n" unless scalar(@names);
@@ -388,7 +386,7 @@ sub cmd_post ($) {
 sub cmd_quit ($) {
 	my ($self) = @_;
 	$self->write(\"205 closing connection - goodbye!\r\n");
-	$self->shutdn;
+	$self->close;
 	undef;
 }
 
@@ -466,17 +464,16 @@ sub set_nntp_headers ($$) {
 
 sub art_lookup ($$$) {
 	my ($self, $art, $code) = @_;
-	my ($ibx, $n);
-	my $err;
+	my ($ibx, $n, $mid, $err);
 	if (defined $art) {
 		if ($art =~ /\A[0-9]+\z/) {
-			$err = \"423 no such article number in this group\r\n";
+			$err = \"423 no such article (#$art) in this group\r\n";
 			$n = int($art);
 			goto find_ibx;
 		} elsif ($art =~ $ONE_MSGID) {
-			($ibx, $n) = mid_lookup($self, $1);
+			($ibx, $n) = mid_lookup($self, $mid = $1);
 			goto found if $ibx;
-			return \r430;
+			return \r430 $mid;
 		} else {
 			return \r501;
 		}
@@ -488,7 +485,13 @@ find_ibx:
 			return \"412 no newsgroup has been selected\r\n";
 	}
 found:
-	my $smsg = $ibx->over(1)->get_art($n) or return $err;
+	my $smsg = $ibx->over(1)->get_art($n) or do {
+		if (defined $mid) {
+			warn "BUG: $ibx->{newsgroup} <$mid> #$n missing";
+			return r430 $mid;
+		}
+		return $err;
+	};
 	$smsg->{-ibx} = $ibx;
 	if ($code == 223) { # STAT
 		set_art($self, $n);
@@ -515,7 +518,7 @@ sub msg_body_write ($$) {
 
 sub set_art {
 	my ($self, $art) = @_;
-	$self->{article} = $art if defined $art && $art =~ /\A[0-9]+\z/;
+	$self->{article} = $art + 0 if defined $art && $art =~ /\A[0-9]+\z/;
 }
 
 sub msg_hdr_write ($$) {
@@ -523,8 +526,7 @@ sub msg_hdr_write ($$) {
 	set_nntp_headers($eml, $smsg);
 
 	my $hdr = $eml->{hdr} // \(my $x = '');
-	# fixup old bug from import (pre-a0c07cba0e5d8b6a)
-	$$hdr =~ s/\A[\r\n]*From [^\r\n]*\r?\n//s;
+	PublicInbox::Eml::strip_from($$hdr);
 	$$hdr =~ s/(?<!\r)\n/\r\n/sg; # Alpine barfs without this
 
 	# for leafnode compatibility, we need to ensure Message-ID headers
@@ -537,7 +539,11 @@ sub blob_cb { # called by git->cat_async via ibx_async_cat
 	my ($bref, $oid, $type, $size, $smsg) = @_;
 	my $self = $smsg->{nntp};
 	my $code = $smsg->{nntp_code};
-	if (!defined($oid)) {
+	if (!defined($type)) {
+		warn "E: git aborted on $oid / $smsg->{blob} ".
+			$self->{-ibx}->{inboxdir};
+		return $self->close;
+	} elsif ($type ne 'blob') {
 		# it's possible to have TOCTOU if an admin runs
 		# public-inbox-(edit|purge), just move onto the next message
 		warn "E: $smsg->{blob} missing in $smsg->{-ibx}->{inboxdir}\n";
@@ -629,8 +635,8 @@ sub hdr_message_id ($$$) { # optimize XHDR Message-ID [range] for slrnpull.
 	my ($self, $xhdr, $range) = @_;
 
 	if (defined $range && $range =~ $ONE_MSGID) {
-		my ($ibx, $n) = mid_lookup($self, $1);
-		return r430 unless $n;
+		my ($ibx, $n) = mid_lookup($self, my $mid = $1);
+		return r430 $mid unless $n;
 		hdr_mid_response($self, $xhdr, $ibx, $n, $range, $range);
 	} else { # numeric range
 		$range = $self->{article} unless defined $range;
@@ -700,7 +706,7 @@ sub hdr_xref ($$$) { # optimize XHDR Xref [range] for rtin
 	if (defined $range && $range =~ $ONE_MSGID) {
 		my $mid = $1;
 		my ($ibx, $n) = mid_lookup($self, $mid);
-		return r430 unless $n;
+		return r430 $mid unless defined $n;
 		my $smsg = $ibx->over(1)->get_art($n) or return;
 		hdr_mid_response($self, $xhdr, $ibx, $n, $range,
 				xref($self, $ibx, $smsg));
@@ -744,8 +750,8 @@ sub smsg_range_i {
 sub hdr_smsg ($$$$) {
 	my ($self, $xhdr, $field, $range) = @_;
 	if (defined $range && $range =~ $ONE_MSGID) {
-		my ($ibx, $n) = mid_lookup($self, $1);
-		return r430 unless defined $n;
+		my ($ibx, $n) = mid_lookup($self, my $mid = $1);
+		return r430 $mid unless defined $n;
 		my $v = over_header_for($ibx, $n, $field);
 		hdr_mid_response($self, $xhdr, $ibx, $n, $range, $v);
 	} else { # numeric range
@@ -846,9 +852,12 @@ sub over_line ($$$) {
 sub cmd_over ($;$) {
 	my ($self, $range) = @_;
 	if ($range && $range =~ $ONE_MSGID) {
-		my ($ibx, $n) = mid_lookup($self, $1);
-		defined $n or return r430;
-		my $smsg = $ibx->over(1)->get_art($n) or return r430;
+		my ($ibx, $n) = mid_lookup($self, my $mid = $1);
+		defined $n or return r430 $mid;
+		my $smsg = $ibx->over(1)->get_art($n) or do {
+			warn "BUG: $ibx->{newsgroup} <$mid> #$n missing";
+			return r430 $mid;
+		};
 		$self->msg_more(
 			"224 Overview information follows (multi-line)\r\n");
 
