@@ -7,10 +7,13 @@ package PublicInbox::Search;
 use strict;
 use v5.10.1;
 use parent qw(Exporter);
+use PublicInbox::Lock;
 our @EXPORT_OK = qw(retry_reopen int_val get_pct xap_terms);
 use List::Util qw(max);
 use POSIX qw(strftime);
+use autodie qw(open pipe);
 use Carp ();
+our $XHC = 0; # defined but false
 
 # values for searching, changing the numeric value breaks
 # compatibility with old indices (so don't change them it)
@@ -35,7 +38,7 @@ use constant {
 	# 4 - change "Re: " normalization, avoid circular Reference ghosts
 	# 5 - subject_path drops trailing '.'
 	# 6 - preserve References: order in document data
-	# 7 - remove references and inreplyto terms
+	# 7 - remove references and inreplyto terms (restored in 15 (v2.0))
 	# 8 - remove redundant/unneeded document data
 	# 9 - disable Message-ID compression (SHA-1)
 	# 10 - optimize doc for NNTP overviews
@@ -52,27 +55,59 @@ use constant {
 	#      * "lid:" and "l:" for List-Id searches
 	#
 	#      v1.6.0 adds BYTES, UID and THREADID values
+	#      v2.0.0 re-adds "references:"
 	SCHEMA_VERSION => 15,
+
+	# we may have up to 8 FDs per shard (depends on Xapian *shrug*)
+	SHARD_COST => 8,
 };
 
 use PublicInbox::Smsg;
-use PublicInbox::Over;
+eval { require PublicInbox::Over };
 our $QP_FLAGS;
 our %X = map { $_ => 0 } qw(BoolWeight Database Enquire QueryParser Stem Query);
-our $Xap; # 'Search::Xapian' or 'Xapian'
+our $Xap; # 'Xapian' or 'Search::Xapian'
 our $NVRP; # '$Xap::'.('NumberValueRangeProcessor' or 'NumberRangeProcessor')
 
 # ENQ_DESCENDING and ENQ_ASCENDING weren't in SWIG Xapian.pm prior to 1.4.16,
 # let's hope the ABI is stable
 our $ENQ_DESCENDING = 0;
 our $ENQ_ASCENDING = 1;
+our @MAIL_VMAP = (
+	[ YYYYMMDD, 'd:', 'YYYYmmdd' ], # enum date_fmt
+	[ TS, 'rt:', 'epoch_sec' ],
+	# these are undocumented for WWW, but lei and IMAP use them
+	[ DT, 'dt:', 'YYYYmmddHHMMSS' ],
+	[ BYTES, 'z:' ],
+	[ UID, 'uid:' ]
+);
+our @MAIL_NRP;
+
+# Getopt::Long spec, only short options for portability in C++ implementation
+our @XH_SPEC = (
+	'a', # ascending sort
+	'c', # code search
+	'd=s@', # shard dirs
+	'g=s', # git dir (with -c)
+	'k=i', # sort column (like sort(1))
+	'l=s', # open.lock path
+	'm=i', # maximum number of results
+	'o=i', # offset
+	'r', # 1=relevance then column
+	't', # collapse threads
+	'u=i', # IMAP UID min
+	'A=s@', # prefixes
+	'K=i', # timeout kill after i seconds
+	'O=s', # eidx_key
+	'T=i', # threadid
+	'U=i', # IMAP UID max
+	'Q=s@', # query prefixes "$user_prefix[:=]$XPREFIX"
+);
 
 sub load_xapian () {
 	return 1 if defined $Xap;
-	# n.b. PI_XAPIAN is intended for development use only.  We still
-	# favor Search::Xapian since that's what's available in current
-	# Debian stable (10.x) and derived distros.
-	for my $x (($ENV{PI_XAPIAN} // 'Search::Xapian'), 'Xapian') {
+	# n.b. PI_XAPIAN is intended for development use only
+	for my $x (($ENV{PI_XAPIAN} // 'Xapian'), 'Search::Xapian') {
 		eval "require $x";
 		next if $@;
 
@@ -85,8 +120,7 @@ sub load_xapian () {
 
 		# NumberRangeProcessor was added in Xapian 1.3.6,
 		# NumberValueRangeProcessor was removed for 1.5.0+,
-		# favor the older /Value/ variant since that's what our
-		# (currently) preferred Search::Xapian supports
+		# continue with the older /Value/ variant for now...
 		$NVRP = $x.'::'.($x eq 'Xapian' && $xver ge v1.5 ?
 			'NumberRangeProcessor' : 'NumberValueRangeProcessor');
 		$X{$_} = $Xap.'::'.$_ for (keys %X);
@@ -101,6 +135,7 @@ sub load_xapian () {
 		# or make indexlevel=medium as default
 		$QP_FLAGS = FLAG_PHRASE() | FLAG_BOOLEAN() | FLAG_LOVEHATE() |
 				FLAG_WILDCARD();
+		@MAIL_NRP = map { $NVRP->new(@$_[0, 1]) } @MAIL_VMAP;
 		return 1;
 	}
 	undef;
@@ -110,43 +145,51 @@ sub load_xapian () {
 # a prefix common in patch emails
 our $LANG = 'english';
 
-# note: the non-X term prefix allocations are shared with
-# Xapian omega, see xapian-applications/omega/docs/termprefixes.rst
-my %bool_pfx_external = (
-	mid => 'Q', # Message-ID (full/exact), this is mostly uniQue
-	lid => 'G', # newsGroup (or similar entity), just inside <>
+our %PATCH_BOOL_COMMON = (
 	dfpre => 'XDFPRE',
 	dfpost => 'XDFPOST',
 	dfblob => 'XDFPRE XDFPOST',
 	patchid => 'XDFID',
 );
 
-my $non_quoted_body = 'XNQ XDFN XDFA XDFB XDFHH XDFCTX XDFPRE XDFPOST XDFID';
-my %prob_prefix = (
-	# for mairix compatibility
+# note: the non-X term prefix allocations are shared with
+# Xapian omega, see xapian-applications/omega/docs/termprefixes.rst
+my %bool_pfx_external = (
+	mid => 'Q', # Message-ID (full/exact), this is mostly uniQue
+	lid => 'G', # newsGroup (or similar entity), just inside <>
+	references => 'XRF',
+	%PATCH_BOOL_COMMON
+);
+
+# for mairix compatibility
+our $NON_QUOTED_BODY = 'XNQ XDFN XDFA XDFB XDFHH XDFCTX XDFPRE XDFPOST XDFID';
+our %PATCH_PROB_COMMON = (
 	s => 'S',
-	m => 'XM', # 'mid:' (bool) is exact, 'm:' (prob) can do partial
-	l => 'XL', # 'lid:' (bool) is exact, 'l:' (prob) can do partial
 	f => 'A',
-	t => 'XTO',
-	tc => 'XTO XCC',
-	c => 'XCC',
-	tcf => 'XTO XCC A',
-	a => 'XTO XCC A',
-	b => $non_quoted_body . ' XQUOT',
-	bs => $non_quoted_body . ' XQUOT S',
+	b => $NON_QUOTED_BODY . ' XQUOT',
+	bs => $NON_QUOTED_BODY . ' XQUOT S',
 	n => 'XFN',
 
 	q => 'XQUOT',
-	nq => $non_quoted_body,
+	nq => $NON_QUOTED_BODY,
 	dfn => 'XDFN',
 	dfa => 'XDFA',
 	dfb => 'XDFB',
 	dfhh => 'XDFHH',
 	dfctx => 'XDFCTX',
+);
 
+my %prob_prefix = (
+	m => 'XM', # 'mid:' (bool) is exact, 'm:' (prob) can do partial
+	l => 'XL', # 'lid:' (bool) is exact, 'l:' (prob) can do partial
+	t => 'XTO',
+	tc => 'XTO XCC',
+	c => 'XCC',
+	tcf => 'XTO XCC A',
+	a => 'XTO XCC A',
+	%PATCH_PROB_COMMON,
 	# default:
-	'' => 'XM S A XQUOT XFN ' . $non_quoted_body,
+	'' => 'XM S A XQUOT XFN ' . $NON_QUOTED_BODY,
 );
 
 # not documenting m: and mid: for now, the using the URLs works w/o Xapian
@@ -154,33 +197,33 @@ my %prob_prefix = (
 # especially since we don't offer boolean searches for To/Cc/From
 # headers, either
 our @HELP = (
-	's:' => 'match within Subject  e.g. s:"a quick brown fox"',
-	'd:' => <<EOF,
+	s => 'match within Subject  e.g. s:"a quick brown fox"',
+	d => <<EOF,
 match date-time range, git "approxidate" formats supported
 Open-ended ranges such as `d:last.week..' and
 `d:..2.days.ago' are supported
 EOF
-	'b:' => 'match within message body, including text attachments',
-	'nq:' => 'match non-quoted text within message body',
-	'q:' => 'match quoted text within message body',
-	'n:' => 'match filename of attachment(s)',
-	't:' => 'match within the To header',
-	'c:' => 'match within the Cc header',
-	'f:' => 'match within the From header',
-	'a:' => 'match within the To, Cc, and From headers',
-	'tc:' => 'match within the To and Cc headers',
-	'l:' => 'match contents of the List-Id header',
-	'bs:' => 'match within the Subject and body',
-	'dfn:' => 'match filename from diff',
-	'dfa:' => 'match diff removed (-) lines',
-	'dfb:' => 'match diff added (+) lines',
-	'dfhh:' => 'match diff hunk header context (usually a function name)',
-	'dfctx:' => 'match diff context lines',
-	'dfpre:' => 'match pre-image git blob ID',
-	'dfpost:' => 'match post-image git blob ID',
-	'dfblob:' => 'match either pre or post-image git blob ID',
-	'patchid:' => "match `git patch-id --stable' output",
-	'rt:' => <<EOF,
+	b => 'match within message body, including text attachments',
+	nq => 'match non-quoted text within message body',
+	q => 'match quoted text within message body',
+	n => 'match filename of attachment(s)',
+	t => 'match within the To header',
+	c => 'match within the Cc header',
+	f => 'match within the From header',
+	a => 'match within the To, Cc, and From headers',
+	tc => 'match within the To and Cc headers',
+	l => 'match contents of the List-Id header',
+	bs => 'match within the Subject and body',
+	dfn => 'match filename from diff',
+	dfa => 'match diff removed (-) lines',
+	dfb => 'match diff added (+) lines',
+	dfhh => 'match diff hunk header context (usually a function name)',
+	dfctx => 'match diff context lines',
+	dfpre => 'match pre-image git blob ID',
+	dfpost => 'match post-image git blob ID',
+	dfblob => 'match either pre or post-image git blob ID',
+	patchid => "match `git patch-id --stable' output",
+	rt => <<EOF,
 match received time, like `d:' if sender's clock was correct
 EOF
 );
@@ -190,33 +233,40 @@ sub xdir ($;$) {
 	my ($self, $rdonly) = @_;
 	if ($rdonly || !defined($self->{shard})) {
 		$self->{xpfx};
-	} else { # v2 + extindex only:
+	} else { # v2, extindex, cindex only:
 		"$self->{xpfx}/$self->{shard}";
 	}
 }
 
-# returns all shards as separate Xapian::Database objects w/o combining
-sub xdb_shards_flat ($) {
+# returns shard directories as an array of strings, does not verify existence
+sub shard_dirs ($) {
 	my ($self) = @_;
 	my $xpfx = $self->{xpfx};
-	my (@xdb, $slow_phrase);
-	load_xapian();
-	$self->{qp_flags} //= $QP_FLAGS;
-	if ($xpfx =~ m!/xapian[0-9]+\z!) {
-		@xdb = ($X{Database}->new($xpfx));
-		$self->{qp_flags} |= FLAG_PHRASE() if !-f "$xpfx/iamchert";
-	} else {
+	if ($xpfx =~ m!/xapian[0-9]+\z!) { # v1 inbox
+		($xpfx);
+	} else { # v2 inbox, eidx, cidx
 		opendir(my $dh, $xpfx) or return (); # not initialized yet
 		# We need numeric sorting so shard[0] is first for reading
 		# Xapian metadata, if needed
 		my $last = max(grep(/\A[0-9]+\z/, readdir($dh))) // return ();
-		for (0..$last) {
-			my $shard_dir = "$self->{xpfx}/$_";
-			push @xdb, $X{Database}->new($shard_dir);
-			$slow_phrase ||= -f "$shard_dir/iamchert";
-		}
-		$self->{qp_flags} |= FLAG_PHRASE() if !$slow_phrase;
+		map { "$xpfx/$_" } (0..$last);
 	}
+}
+
+sub open_lock ($) { "$_[0]->{xpfx}/../open.lock" }
+
+# returns all shards as separate Xapian::Database objects w/o combining
+sub xdb_shards_flat ($) {
+	my ($self) = @_;
+	load_xapian();
+	$self->{qp_flags} //= $QP_FLAGS;
+	my $slow_phrase;
+	my $lk = PublicInbox::Lock::may_sh open_lock($self);
+	my @xdb = map {
+		$slow_phrase ||= -f "$_/iamchert";
+		$X{Database}->new($_); # raises if missing
+	} shard_dirs($self);
+	$self->{qp_flags} |= FLAG_PHRASE() if !$slow_phrase;
 	@xdb;
 }
 
@@ -229,6 +279,12 @@ sub mdocid {
 	int(($docid - 1) / $nshard) + 1;
 }
 
+sub docids_to_artnums {
+	my $nshard = shift->{nshard};
+	# XXX does array vs arrayref make a difference in modern Perls?
+	map { int(($_ - 1) / $nshard) + 1 } @_;
+}
+
 sub mset_to_artnums {
 	my ($self, $mset) = @_;
 	my $nshard = $self->{nshard};
@@ -246,19 +302,34 @@ sub xdb ($) {
 	};
 }
 
+sub load_extra_indexers ($$) {
+	my ($self, $ibx) = @_;
+	my @extra;
+	for my $f (qw(IndexHeader AltId)) {
+		my $specs = $ibx->{lc $f} // next;
+		my $cls = "PublicInbox::$f";
+		eval "require $cls" or die $@;
+		push @extra, map { $cls->new($ibx, $_) } @$specs;
+	}
+	$self->{-extra} = \@extra if @extra;
+}
+
 sub new {
 	my ($class, $ibx) = @_;
 	ref $ibx or die "BUG: expected PublicInbox::Inbox object: $ibx";
 	my $xap = $ibx->version > 1 ? 'xap' : 'public-inbox/xapian';
 	my $xpfx = "$ibx->{inboxdir}/$xap".SCHEMA_VERSION;
 	my $self = bless { xpfx => $xpfx }, $class;
-	$self->{altid} = $ibx->{altid} if defined($ibx->{altid});
+	$self->load_extra_indexers($ibx);
 	$self;
 }
 
 sub reopen {
 	my ($self) = @_;
 	if (my $xdb = $self->{xdb}) {
+		my $lk = defined($self->{xpfx}) ?
+			PublicInbox::Lock::may_sh(open_lock($self)) :
+			undef;
 		$xdb->reopen;
 	}
 	$self; # make chaining easier
@@ -267,8 +338,8 @@ sub reopen {
 # Convert git "approxidate" ranges to something usable with our
 # Xapian indices.  At the moment, Xapian only offers a C++-only API
 # and neither the SWIG nor XS bindings allow us to use custom code
-# to parse dates (and libgit2 doesn't expose git__date_parse, either,
-# so we're running git-rev-parse(1)).
+# to parse dates (and libgit2 doesn't expose git_date_parse publicly,
+# either, so we're running git-rev-parse(1)).
 # This replaces things we need to send to $git->date_parse with
 # "\0".$strftime_format.['+'|$idx]."\0" placeholders
 sub date_parse_prepare {
@@ -277,39 +348,22 @@ sub date_parse_prepare {
 	my $end = $range =~ s/([\)\s]*)\z// ? $1 : '';
 	my @r = split(/\.\./, $range, 2);
 
-	# expand "d:20101002" => "d:20101002..20101003" and like
+	# expand "dt:2010-10-02" => "dt:2010-10-02..2010-10-03" and like
 	# n.b. git doesn't do YYYYMMDD w/o '-', it needs YYYY-MM-DD
-	# We upgrade "d:" to "dt:" to iff using approxidate
 	if ($pfx eq 'd') {
-		my $fmt = "\0%Y%m%d";
-		if (!defined($r[1])) {
+		if (!defined($r[1])) { # git needs gaps and not /\d{14}/
 			if ($r[0] =~ /\A([0-9]{4})([0-9]{2})([0-9]{2})\z/) {
-				push @$to_parse, "$1-$2-$3";
-				# we could've handled as-is, but we need
-				# to parse anyways for "d+" below
+				push @$to_parse, "$1-$2-$3 00:00:00";
 			} else {
 				push @$to_parse, $r[0];
-				if ($r[0] !~ /\A[0-9]{4}-[0-9]{2}-[0-9]{2}\z/) {
-					$pfx = 'dt';
-					$fmt = "\0%Y%m%d%H%M%S";
-				}
 			}
-			$r[0] = "$fmt+$#$to_parse\0";
-			$r[1] = "$fmt+\0";
+			$r[0] = "\0%Y%m%d$#$to_parse\0";
+			$r[1] = "\0%Y%m%d+\0";
 		} else {
-			for my $x (@r) {
-				next if $x eq '' || $x =~ /\A[0-9]{8}\z/;
-				push @$to_parse, $x;
-				if ($x !~ /\A[0-9]{4}-[0-9]{2}-[0-9]{2}\z/) {
-					$pfx = 'dt';
-				}
-				$x = "$fmt$#$to_parse\0";
-			}
-			if ($pfx eq 'dt') {
-				for (@r) {
-					s/\0%Y%m%d/\0%Y%m%d%H%M%S/;
-					s/\A([0-9]{8})\z/${1}000000/;
-				}
+			for (@r) {
+				next if $_ eq '' || /\A[0-9]{8}\z/;
+				push @$to_parse, $_;
+				$_ = "\0%Y%m%d$#$to_parse\0";
 			}
 		}
 	} elsif ($pfx eq 'dt') {
@@ -329,7 +383,7 @@ sub date_parse_prepare {
 				$x = "\0%Y%m%d%H%M%S$#$to_parse\0";
 			}
 		}
-	} else { # "rt", let git interpret "YYYY", deal with Y10K later :P
+	} else { # (rt|ct), let git interpret "YYYY", deal with Y10K later :P
 		for my $x (@r) {
 			next if $x eq '' || $x =~ /\A[0-9]{5,}\z/;
 			push @$to_parse, $x;
@@ -357,6 +411,7 @@ sub date_parse_finalize {
 # if this $argv isn't from a command execution
 sub query_argv_to_string {
 	my (undef, $git, $argv) = @_;
+	return "@$argv" if $XHC && $XHC->{impl} eq 'PublicInbox::XapHelperCxx';
 	my $to_parse;
 	my $tmp = join(' ', map {;
 		if (s!\b(d|rt|dt):(\S+)\z!date_parse_prepare(
@@ -375,6 +430,7 @@ sub query_argv_to_string {
 # this is for the WWW "q=" query parameter and "lei q --stdin"
 # it can't do d:"5 days ago", but it will do d:5.days.ago
 sub query_approxidate {
+	return if $XHC && $XHC->{impl} eq 'PublicInbox::XapHelperCxx';
 	my (undef, $git) = @_; # $_[2] = $query_string (modified in-place)
 	my $DQ = qq<"\x{201c}\x{201d}>; # Xapian can use curly quotes
 	$_[2] =~ tr/\x00/ /; # Xapian doesn't do NUL, we use it as a placeholder
@@ -388,13 +444,98 @@ sub query_approxidate {
 	date_parse_finalize($git, $to_parse, $_[2]) if $to_parse;
 }
 
-# read-only
+# read-only, for mail only (codesearch has different rules)
 sub mset {
-	my ($self, $query_string, $opts) = @_;
-	$opts ||= {};
-	my $qp = $self->{qp} //= $self->qparse_new;
-	my $query = $qp->parse_query($query_string, $self->{qp_flags});
-	_do_enquire($self, $query, $opts);
+	my ($self, $qry_str, $opt) = @_;
+	my $qp = $self->{qp} // $self->qparse_new;
+	my $qry = $qp->parse_query($qry_str, $self->{qp_flags});
+	if (defined(my $eidx_key = $opt->{eidx_key})) {
+		$qry = $X{Query}->new(OP_FILTER(), $qry, 'O'.$eidx_key);
+	}
+	if (defined(my $uid_range = $opt->{uid_range})) {
+		my $range = $X{Query}->new(OP_VALUE_RANGE(), UID,
+					sortable_serialise($uid_range->[0]),
+					sortable_serialise($uid_range->[1]));
+		$qry = $X{Query}->new(OP_FILTER(), $qry, $range);
+	}
+	if (defined(my $tid = $opt->{threadid})) {
+		$tid = sortable_serialise($tid);
+		$qry = $X{Query}->new(OP_FILTER(), $qry,
+			$X{Query}->new(OP_VALUE_RANGE(), THREADID, $tid, $tid));
+	}
+	$opt->{sort_col} //= TS;
+	do_enquire($self, $qry, $opt);
+}
+
+my %QPMETHOD_2_SYM = (add_prefix => ':', add_boolean_prefix => '=');
+
+sub xh_opt ($$) {
+	my ($self, $opt) = @_;
+	my $lim = $opt->{limit} || 50;
+	my @ret;
+	push @ret, '-o', $opt->{offset} if $opt->{offset};
+	push @ret, '-m', $lim;
+	push @ret, '-r' if $opt->{relevance};
+	push @ret, '-k', $opt->{sort_col} // TS;
+	push @ret, '-a' if $opt->{asc};
+	push @ret, '-t' if $opt->{threads};
+	push @ret, '-T', $opt->{threadid} if defined $opt->{threadid};
+	push @ret, '-O', $opt->{eidx_key} if defined $opt->{eidx_key};
+	if (my $uid = $opt->{uid_range}) {
+		push @ret, '-u', $uid->[0], '-U', $uid->[1];
+	}
+	@ret;
+}
+
+# returns a true value if actually handled asynchronously,
+# and a falsy value if handled synchronously
+sub async_mset {
+	my ($self, $qry, $opt, $cb, @args) = @_;
+	if ($XHC) { # unconditionally retrieving pct + rank for now
+		xdb($self); # populate {nshards}
+		my @margs = ($self->xh_args, xh_opt($self, $opt), '--');
+		my $ret = eval {
+			pipe my $out_rd, my $out_wr;
+			open my $err_rw, '+>', undef;
+			$XHC->mkreq([ $out_wr, $err_rw ], 'mset', @margs,
+				(ref($qry) eq 'ARRAY' ? @$qry : $qry));
+			undef $out_wr;
+			PublicInbox::XhcMset->maybe_new($out_rd, $err_rw,
+							$self, $cb, @args);
+		};
+		$cb->(@args, undef, $@) if $@;
+		$ret;
+	} else { # synchronous
+		my $mset;
+		for my $qstr (ref($qry) eq 'ARRAY' ? @$qry : ($qry)) {
+			$mset = eval { $self->mset($qstr, $opt) };
+			last if $mset && $mset->size;
+		}
+		$mset ? $cb->(@args, $mset) : $cb->(@args, undef, $@ || 'err');
+		undef;
+	}
+}
+
+sub do_enquire { # shared with CodeSearch and MiscSearch
+	my ($self, $qry, $opt) = @_;
+	my $enq = $X{Enquire}->new(xdb($self));
+	$enq->set_query($qry);
+	my $col = $opt->{sort_col} // TS;
+	if ($col < 0) {
+		$enq->set_weighting_scheme($X{BoolWeight}->new);
+		$enq->set_docid_order(
+			$opt->{asc} ? $ENQ_ASCENDING : $ENQ_DESCENDING);
+	} elsif ($opt->{relevance}) {
+		$enq->set_sort_by_relevance_then_value($col, !$opt->{asc});
+	} else {
+		$enq->set_sort_by_value_then_relevance($col, !$opt->{asc});
+	}
+	# `lei q -t / --threads' or JMAP collapseThreads; but don't collapse
+	# on `-tt' ({threads} > 1) which sets the Flagged|Important keyword
+	(($opt->{threads} // 0) == 1 && has_threadid($self)) and
+		$enq->set_collapse_key(THREADID);
+	retry_reopen($self, \&enquire_once, $enq,
+			$opt->{offset} || 0, $opt->{limit} || 50);
 }
 
 sub retry_reopen {
@@ -421,50 +562,15 @@ sub retry_reopen {
 	Carp::croak("Too many Xapian database modifications in progress\n");
 }
 
-sub _do_enquire {
-	my ($self, $query, $opts) = @_;
-	retry_reopen($self, \&_enquire_once, $query, $opts);
-}
-
 # returns true if all docs have the THREADID value
 sub has_threadid ($) {
 	my ($self) = @_;
 	(xdb($self)->get_metadata('has_threadid') // '') eq '1';
 }
 
-sub _enquire_once { # retry_reopen callback
-	my ($self, $query, $opts) = @_;
-	my $xdb = xdb($self);
-	if (defined(my $eidx_key = $opts->{eidx_key})) {
-		$query = $X{Query}->new(OP_FILTER(), $query, 'O'.$eidx_key);
-	}
-	if (defined(my $uid_range = $opts->{uid_range})) {
-		my $range = $X{Query}->new(OP_VALUE_RANGE(), UID,
-					sortable_serialise($uid_range->[0]),
-					sortable_serialise($uid_range->[1]));
-		$query = $X{Query}->new(OP_FILTER(), $query, $range);
-	}
-	my $enquire = $X{Enquire}->new($xdb);
-	$enquire->set_query($query);
-	$opts ||= {};
-	my $rel = $opts->{relevance} // 0;
-	if ($rel == -2) { # ORDER BY docid/UID (highest first)
-		$enquire->set_weighting_scheme($X{BoolWeight}->new);
-		$enquire->set_docid_order($ENQ_DESCENDING);
-	} elsif ($rel == -1) { # ORDER BY docid/UID (lowest first)
-		$enquire->set_weighting_scheme($X{BoolWeight}->new);
-		$enquire->set_docid_order($ENQ_ASCENDING);
-	} elsif ($rel == 0) {
-		$enquire->set_sort_by_value_then_relevance(TS, !$opts->{asc});
-	} else { # rel > 0
-		$enquire->set_sort_by_relevance_then_value(TS, !$opts->{asc});
-	}
-
-	# `mairix -t / --threads' or JMAP collapseThreads
-	if ($opts->{threads} && has_threadid($self)) {
-		$enquire->set_collapse_key(THREADID);
-	}
-	$enquire->get_mset($opts->{offset} || 0, $opts->{limit} || 50);
+sub enquire_once { # retry_reopen callback
+	my (undef, $enq, $offset, $limit) = @_;
+	$enq->get_mset($offset, $limit);
 }
 
 sub mset_to_smsg {
@@ -481,66 +587,112 @@ sub mset_to_smsg {
 # read-write
 sub stemmer { $X{Stem}->new($LANG) }
 
-# read-only
-sub qparse_new {
+sub qp_init_common {
 	my ($self) = @_;
-
-	my $xdb = xdb($self);
-	my $qp = $X{QueryParser}->new;
+	my $qp = $self->{qp} = $X{QueryParser}->new;
 	$qp->set_default_op(OP_AND());
-	$qp->set_database($xdb);
+	$qp->set_database(xdb($self));
 	$qp->set_stemmer(stemmer($self));
 	$qp->set_stemming_strategy(STEM_SOME());
 	my $cb = $qp->can('set_max_wildcard_expansion') //
 		$qp->can('set_max_expansion'); # Xapian 1.5.0+
 	$cb->($qp, 100);
-	$cb = $qp->can('add_valuerangeprocessor') //
+	$qp;
+}
+
+# read-only
+sub qparse_new {
+	my ($self) = @_;
+	my $qp = qp_init_common($self);
+	my $cb = $qp->can('add_valuerangeprocessor') //
 		$qp->can('add_rangeprocessor'); # Xapian 1.5.0+
-	$cb->($qp, $NVRP->new(YYYYMMDD, 'd:'));
-	$cb->($qp, $NVRP->new(DT, 'dt:'));
-
-	# for IMAP, undocumented for WWW and may be split off go away
-	$cb->($qp, $NVRP->new(BYTES, 'z:'));
-	$cb->($qp, $NVRP->new(TS, 'rt:'));
-	$cb->($qp, $NVRP->new(UID, 'uid:'));
 
+	$cb->($qp, $_) for @MAIL_NRP;
 	while (my ($name, $prefix) = each %bool_pfx_external) {
 		$qp->add_boolean_prefix($name, $_) foreach split(/ /, $prefix);
 	}
 
-	# we do not actually create AltId objects,
-	# just parse the spec to avoid the extra DB handles for now.
-	if (my $altid = $self->{altid}) {
-		my $user_pfx = $self->{-user_pfx} = [];
-		for (@$altid) {
-			# $_ = 'serial:gmane:/path/to/gmane.msgmap.sqlite3'
-			# note: Xapian supports multibyte UTF-8, /^[0-9]+$/,
-			# and '_' with prefixes matching \w+
-			/\Aserial:(\w+):/ or next;
-			my $pfx = $1;
-			push @$user_pfx, "$pfx:", <<EOF;
-alternate serial number  e.g. $pfx:12345 (boolean)
-EOF
-			# gmane => XGMANE
-			$qp->add_boolean_prefix($pfx, 'X'.uc($pfx));
-		}
-		chomp @$user_pfx;
+	for my $x (@{$self->{-extra} // []}) {
+		my $m = $x->query_parser_method;
+		$qp->$m(@$x{qw(prefix xprefix)});
 	}
-
 	while (my ($name, $prefix) = each %prob_prefix) {
 		$qp->add_prefix($name, $_) foreach split(/ /, $prefix);
 	}
 	$qp;
 }
 
-sub help {
-	my ($self) = @_;
-	$self->{qp} //= $self->qparse_new; # parse altids
-	my @ret = @HELP;
-	if (my $user_pfx = $self->{-user_pfx}) {
-		push @ret, @$user_pfx;
+sub generate_cxx () { # generates snippet for xap_helper.h
+	my $gdfp_size = grep { @$_ == 3 } @MAIL_VMAP;
+	my @gdfp_pfx;
+	my $ret = <<EOM;
+# line ${\__LINE__} "${\__FILE__}"
+static RP *mail_rp[${\scalar(@MAIL_VMAP)}];
+static GitDateFieldProcessor *mail_gdfp[$gdfp_size];
+static void mail_rp_init(void)
+{
+EOM
+	for (0..$#MAIL_VMAP) {
+		my $x = $MAIL_VMAP[$_];
+		if (@$x == 3) {
+			push @gdfp_pfx, $x->[1];
+			$ret .= <<"";
+	mail_rp[$_] = new GitDateRangeProcessor($x->[0], "$x->[1]", $x->[2]);
+	mail_gdfp[$#gdfp_pfx] = new GitDateFieldProcessor($x->[0], $x->[2]);
+
+		} else {
+			$ret .= <<"";
+	mail_rp[$_] = new NRP($x->[0], "$x->[1]");
+
+		}
 	}
-	\@ret;
+$ret .= <<EOM;
+}
+
+# line ${\__LINE__} "${\__FILE__}"
+static void qp_init_mail_search(Xapian::QueryParser *qp)
+{
+	for (size_t i = 0; i < MY_ARRAY_SIZE(mail_rp); i++)
+		qp->ADD_RP(mail_rp[i]);
+EOM
+	for my $name (sort keys %bool_pfx_external) {
+		for (split(/ /, $bool_pfx_external{$name})) {
+			$ret .= qq{\tqp->add_boolean_prefix("$name", "$_");\n}
+		}
+	}
+	my $i = 0;
+	for (@gdfp_pfx) {
+		$ret .= qq{\tqp->add_boolean_prefix("$_", mail_gdfp[$i]);\n};
+		++$i;
+	}
+	# altid support is handled in xh_opt and srch_init_extra in XH
+	for my $name (sort keys %prob_prefix) {
+		for (split(/ /, $prob_prefix{$name})) {
+			$ret .= qq{\tqp->add_prefix("$name", "$_");\n}
+		}
+	}
+	$ret .= "}\n";
+}
+
+sub help2txt (@) { # also used by Documentation/common.perl
+	my @help = @_;
+	my $pad = 0;
+	my $htxt = '';
+	while (defined(my $pfx = shift @help)) {
+		my $n = length($pfx) + 1;
+		$pad = $n if $n > $pad;
+		$htxt .= $pfx . ":\0" . shift(@help) . "\f\n";
+	}
+	$pad += 2;
+	my $padding = ' ' x ($pad + 4);
+	$htxt =~ s/^/$padding/gms;
+	$htxt =~ s/^$padding(\S+)\0/"    $1".(' ' x ($pad - length($1)))/egms;
+	$htxt =~ s/\f\n/\n/gs;
+	$htxt;
+}
+
+sub help_txt {
+	help2txt(@HELP, map { $_->user_help } @{$_[0]->{-extra} // []});
 }
 
 # always returns a scalar value
@@ -560,24 +712,83 @@ sub get_pct ($) { # mset item
 
 sub xap_terms ($$;@) {
 	my ($pfx, $xdb_or_doc, @docid) = @_; # @docid may be empty ()
-	my %ret;
 	my $end = $xdb_or_doc->termlist_end(@docid);
 	my $cur = $xdb_or_doc->termlist_begin(@docid);
+	$cur->skip_to($pfx);
+	my (@ret, $tn);
+	my $pfxlen = length($pfx);
 	for (; $cur != $end; $cur++) {
-		$cur->skip_to($pfx);
-		last if $cur == $end;
-		my $tn = $cur->get_termname;
-		$ret{substr($tn, length($pfx))} = undef if !index($tn, $pfx);
+		$tn = $cur->get_termname;
+		index($tn, $pfx) ? last : push(@ret, substr($tn, $pfxlen));
 	}
-	wantarray ? sort(keys(%ret)) : \%ret;
+	wantarray ? @ret : +{ map { $_ => undef } @ret };
 }
 
 # get combined docid from over.num:
-# (not generic Xapian, only works with our sharding scheme)
+# (not generic Xapian, only works with our sharding scheme for mail)
 sub num2docid ($$) {
 	my ($self, $num) = @_;
 	my $nshard = $self->{nshard};
 	($num - 1) * $nshard + $num % $nshard + 1;
 }
 
+sub all_terms {
+	my ($self, $pfx) = @_;
+	my $cur = xdb($self)->allterms_begin($pfx);
+	my $end = $self->{xdb}->allterms_end($pfx);
+	my $pfxlen = length($pfx);
+	my @ret;
+	for (; $cur != $end; $cur++) {
+		push @ret, substr($cur->get_termname, $pfxlen);
+	}
+	wantarray ? @ret : +{ map { $_ => undef } @ret };
+}
+
+sub xh_args { # prep getopt args to feed to xap_helper.h socket
+	my ($self) = @_;
+	my $apfx = $self->{-alt_pfx} //= do {
+		my %dedupe;
+		for my $x (@{$self->{-extra} // []}) {
+			my $sym = $QPMETHOD_2_SYM{$x->query_parser_method};
+			$dedupe{$x->{prefix}.$sym.$x->{xprefix}} = undef;
+		}
+		# TODO: arbitrary header indexing goes here
+		my @apfx = map { ('-Q', $_) } sort keys %dedupe;
+		\@apfx;
+	};
+	((map { ('-d', $_) } $self->shard_dirs), $self->xh_lock_args, @$apfx);
+}
+
+sub xh_lock_args { ('-l', open_lock($_[0])) }
+
+sub docids_by_postlist ($$) {
+	my ($self, $q) = @_;
+	my $cur = $self->xdb->postlist_begin($q);
+	my $end = $self->{xdb}->postlist_end($q);
+	my @ids;
+	for (; $cur != $end; $cur++) { push(@ids, $cur->get_docid) };
+	@ids;
+}
+
+sub get_doc ($$) {
+	my ($self, $docid) = @_;
+	eval { $self->{xdb}->get_document($docid) } // do {
+		die $@ if $@ && ref($@) !~ /\bDocNotFoundError\b/;
+		undef;
+	}
+}
+
+# not sure where best to put this...
+sub ulimit_n () {
+	my $n;
+	if (eval { require BSD::Resource; 1 }) {
+		my $NOFILE = BSD::Resource::RLIMIT_NOFILE();
+		($n, undef) = BSD::Resource::getrlimit($NOFILE);
+	} else {
+		require PublicInbox::Spawn;
+		$n = PublicInbox::Spawn::run_qx([qw(/bin/sh -c), 'ulimit -n']);
+	}
+	$n;
+}
+
 1;
