@@ -146,6 +146,7 @@ sub cb_spawn ($$$$) {
 	my ($cb, $ibxish, $args, $opt) = @_; # $cb = cpdb() or compact()
 	my $pid = PublicInbox::DS::fork_persist;
 	return $pid if $pid > 0;
+	$SIG{PIPE} = 'DEFAULT'; # warn may fail
 	$SIG{__DIE__} = sub { warn @_; _exit(1) }; # don't jump up stack
 	$cb->($ibxish, $args, $opt);
 	_exit(0);
@@ -276,9 +277,9 @@ sub prepare_run {
 		if ($opt->{cow}) { # make existing $DIR/{xap,ei}* CoW
 			my $dfd = dup(fileno($dh)) // die "dup: $!";
 			open my $fh, '<&='.$dfd;
-			closedir $dh;
 			PublicInbox::Syscall::yesdatacow_fh($fh);
 		}
+		closedir $dh;
 		die "No Xapian shards found in $old\n" unless @old_shards;
 		@old_shards = sort { $a <=> $b } @old_shards;
 		my ($src, $max_shard);
@@ -472,6 +473,7 @@ sub xapian_write_prep ($) {
 	my $flag = eval($PublicInbox::Search::Xap.'::DB_CREATE()');
 	die if $@;
 	$flag |= $PublicInbox::SearchIdx::DB_NO_SYNC if !$opt->{fsync};
+	$flag |= $PublicInbox::SearchIdx::DB_DANGEROUS;
 	(\%PublicInbox::Search::X, $flag);
 }
 
@@ -597,11 +599,15 @@ sub cpdb ($$$) { # cb_spawn callback
 			my $lc = $src->get_metadata('last_commit');
 			$dst->set_metadata('last_commit', $lc) if $lc;
 
-			# only the first xapian shard (0) gets 'indexlevel'
+			# only the first xapian shard (0) gets metadata
 			if ($new =~ m!/(?:xapian[0-9]+|(?:ei|xap)[0-9]+/0)\b!) {
 				my $l = $src->get_metadata('indexlevel');
 				$l eq 'medium' and
 					$dst->set_metadata('indexlevel', $l);
+				for my $k (qw(has_threadid skip_docdata)) {
+					my $v = $src->get_metadata($k);
+					$dst->set_metadata($k, $v) if $v;
+				}
 			}
 			if ($pr_data) {
 				my $tot = $src->get_doccount;
